"use strict";(self.webpackChunkipycanvas=self.webpackChunkipycanvas||[]).push([[509],{509:(a,e,n)=>{n.r(e),n.d(e,{AsyncValueWidgetModel:()=>d.AsyncValueWidgetModel,CanvasManagerModel:()=>d.CanvasManagerModel,CanvasModel:()=>d.CanvasModel,CanvasView:()=>d.CanvasView,LinearGradientModel:()=>d.LinearGradientModel,MODULE_NAME:()=>l.p,MODULE_VERSION:()=>l.e,MultiCanvasModel:()=>d.MultiCanvasModel,MultiCanvasView:()=>d.MultiCanvasView,OffscreenCanvasModel:()=>s.OffscreenCanvasModel,OffscreenCanvasView:()=>s.OffscreenCanvasView,Path2DModel:()=>d.Path2DModel,PatternModel:()=>d.PatternModel,RadialGradientModel:()=>d.RadialGradientModel,RoughCanvasModel:()=>d.RoughCanvasModel});var l=n(797),d=n(815),s=n(836)}}]);